# -*- perl -*-

#
# $Id: GfxConvert.pm,v 1.1 1998/06/25 18:23:45 eserte Exp $
# Author: Slaven Rezic
#
# Copyright (C) 1998 Slaven Rezic. All rights reserved.
# This package is free software; you can redistribute it and/or
# modify it under the same terms as Perl itself.
#
# Mail: eserte@cs.tu-berlin.de
# WWW:  http://user.cs.tu-berlin.de/~eserte/
#

package GfxConvert;

use BBBikeUtil;
use strict;
use vars qw(@tmpfiles $VERBOSE);

my %convsub =
  ('ps' => { 'ppm'  => \&ps2ppm,
	     'gif'  => \&ps2gif,
	     'jpeg' => \&ps2jpeg,
	   }
  );

sub convert {
    my($infmt, $outfmt, $infile, $outfile, %args) = @_;
    my $convsub = $convsub{$infmt}->{$outfmt};
    if (defined $convsub) {
	$convsub->($infile, $outfile, %args);
    } else {
	die "Konversion von $infmt nach $outfmt kann nicht durchgefhrt werden.";
    }
}

sub ps2ppm {
    my($infile, $outfile, %args) = @_;
    if (!is_in_path("gs")) {
	die "Ghostscript wird bentigt.";
    }
    my(@cmd) = (qw(gs -q -sDEVICE=ppmraw -DNOPAUSE));
    if ($args{-res}) {
	push @cmd, "-r$args{-res}";
    }
    push @cmd, ("-sOutputFile=$outfile",
		qw(--),
		$infile);
    warn "Executing @cmd ..." if $VERBOSE;
    if (system(@cmd) != 0) {
	die "Konvertierung mit gs fehlgeschlagen (exit: $?)";
    }
    if (defined $args{'-rotate'}) {
	if (!is_in_path("pnmrotate")) {
	    warn "Warnung: pnmrotate ist nicht vorhanden, keine Umwandlung nach Landscape.";
	} else {
	    my $tmpfile = "/tmp/GfxConvert-ps2ppm.$$.ppm";
	    push @tmpfiles, $tmpfile;
	    my $cmd = "pnmrotate $args{'-rotate'} $outfile > $tmpfile";
	    warn "Executing $cmd ..." if $VERBOSE;
	    if (system($cmd) != 0) {
		warn "Warnung: Fehler beim Aufruf von pnmrotate.";
	    } else {
		warn "Mv from $tmpfile to $outfile ..." if $VERBOSE;
		require File::Copy;
		File::Copy::mv($tmpfile, $outfile);
	    }
	}
    }
    if (defined $args{'-mapcolor'}) {
	if (!is_in_path("ppmchange")) {
	    warn "Warnung: ppmchange ist nicht vorhanden, keine Anpassung der Farben.";
	} else {
	    my $tmpfile = "/tmp/GfxConvert-ps2ppm.$$.ppm";
	    push @tmpfiles, $tmpfile;
	    # XXX escaping mit \ sollte nicht ntig sein!
	    my $arg = join(" ", map { "\\$_ \\$args{'-mapcolor'}->{$_}" }
			   keys %{ $args{'-mapcolor'} });
	    my $cmd = "ppmchange $arg $outfile > $tmpfile";
	    warn "Executing $cmd ..." if $VERBOSE;
	    if (system($cmd) != 0) {
		warn "Warnung: Fehler beim Aufruf von ppmchange.";
	    } else {
		warn "Mv from $tmpfile to $outfile ..." if $VERBOSE;
		require File::Copy;
		File::Copy::mv($tmpfile, $outfile);
	    }
	}
    }
    1;
}

sub ps2gif {
    my($infile, $outfile, %args) = @_;
    if (!is_in_path("ppmtogif")) {
	die "ppmtogif aus der netpbm-Distribution wird bentigt.";
    }
    my $ppmfile = "/tmp/GfxConvert.$$.ppm";
    push @tmpfiles, $ppmfile;
    ps2ppm($infile, $ppmfile, %args);
    my $cmd = "ppmtogif $ppmfile > $outfile";
    warn "Executing $cmd ..." if $VERBOSE;
    if (system($cmd) != 0) {
	die "Konvertierung mit ppmtogif fehlgeschlagen (exit: $?)";
    }
    1;
}

sub ps2jpeg {
    my($infile, $outfile, %args) = @_;
    if (!is_in_path("cjpeg")) {
	die "cjpeg aus der JPEG-Distribution wird bentigt.";
    }
    my $ppmfile = "/tmp/GfxConvert.$$.ppm";
    push @tmpfiles, $ppmfile;
    ps2ppm($infile, $ppmfile, %args);
    my $cmd = "cjpeg -quality 70 $ppmfile > $outfile";
    warn "Executing $cmd ..." if $VERBOSE;
    if (system($cmd) != 0) {
	die "Konvertierung mit cjpeg fehlgeschlagen (exit: $?)";
    }
    1;
}

1;
