# -*- perl -*-

#
# $Id: Strassen.pm,v 2.4 1998/08/01 21:10:44 eserte Exp $
# Author: Slaven Rezic (eserte@cs.tu-berlin.de)
#

# XXX
# Methoden, die ein neues Strassen-Interface bentigt
# next() liefert ein Strasse-Objekt (siehe miscsrc/convbin.pl)
# Attribute: Einbahnstrae, Farbe, Stipple, Outline, Straenoberflche,
#            Eignung fr Radfahrer (?)
# Umbenennen, weil "Strassen" auch Flchen sein knnen (Name?)

use strict;
use FindBin;
use lib ("$FindBin::RealBin/lib", "/home/e/eserte/lib/perl");

######################################################################
package Strassen::Util;
use BBBikeUtil;
#use AutoLoader 'AUTOLOAD';
use vars qw($VERBOSE $tmpdir
	    $cachedir $cacheable @cacheable %cache_symbol
	    $datadumper_var $use_virt_array);

$tmpdir = $ENV{TMPDIR} || $ENV{TMP} || "/tmp";
if (!-d $tmpdir || !-w $tmpdir) { undef $tmpdir }
$cachedir = (-d "$FindBin::RealBin/cache"
	     ? "$FindBin::RealBin/cache"
	     : $tmpdir);
$cacheable = 1;
@cacheable = qw(VirtArray Storable Data::Dumper);
$use_virt_array = 1;

# Argument: [x1,y1], [x2, y2]
sub strecke {
    CORE::sqrt(sqr($_[0]->[0] - $_[1]->[0]) +
	       sqr($_[0]->[1] - $_[1]->[1]));
}

# Argument: "x1,y1", "x2,y2"
sub strecke_s {
    my($x1, $y1) = split(/,/, $_[0]);
    my($x2, $y2) = split(/,/, $_[1]);
    CORE::sqrt(sqr($x1-$x2) + sqr($y1-$y2));
}

# Argumente: Indices
sub strecke_i {
    my($self, $i1, $i2) = @_;
    my($x1,$y1) = unpack("l2", $self->{Index2Coord}[$i1]);
    my($x2,$y2) = unpack("l2", $self->{Index2Coord}[$i2]);
    CORE::sqrt(sqr($x1-$x2) + sqr($y1-$y2));
}

#sub sqr { $_[0] * $_[0] }

sub try_cache {
    my($filename, $write, $ref) = @_;
    my $cache_type;
    my $rw_text = ($write ? 'write' : 'read');
    foreach $cache_type (@cacheable) {
	warn "$rw_text $filename: trying cache type $cache_type...\n"
	  if $VERBOSE;
	if ($cache_type eq 'VirtArray') {
	    # soll VirtArray verwendet werden?
	    next unless $use_virt_array;
	    # wird berhaupt ein Array gespeichert?
	    next if defined $ref && ref $ref ne 'ARRAY';
	    eval q{ local $SIG{'__DIE__'};
		    require VirtArray;
		    die unless VirtArray->can("TIEHASH");
		};
	    if (!$@) {
		if (!$write) {
		    # ist die existierende Datei berhaupt
		    # eine VirtArray-Datei?
		    next if !VirtArray::is_valid($filename);
		}
		return ($cache_type, $filename);
	    }
	} elsif ($cache_type eq 'Storable') {
	    eval q{ local $SIG{'__DIE__'};
		    require Storable;
		    Storable->VERSION(0.509);
		};
	    if (!$@) {
		next if !$write && !-f $filename; # Datei existiert nicht
		return ($cache_type, $filename);
	    }
	} elsif ($cache_type eq 'Data::Dumper') {
	    if ($write) {
		eval q{ local $SIG{'__DIE__'};
			require Data::Dumper;
			Data::Dumper->VERSION(2.10);
		    };
		if (!$@) {
		    return ($cache_type, "$filename.pl");
		}
	    } else {
		if (-f "$filename.pl") {
		    return ($cache_type, "$filename.pl");
		}
	    }
	} else {
	    die "Unknown cache type $cache_type";
	}
    }
    undef;
}

# Gibt wahr zurck, wenn der Cache nicht lter als alle Source-Dateien
# in $srcref (Array-Referenz oder ein String) ist.
sub valid_cache {
    my($cachefile, $srcref) = @_;

    my(@stat_cache) = stat $cachefile;
    if (!defined $stat_cache[9]) { # cachefile nicht vorhanden
	return undef;
    }
    my $src;
    my(@src) = (ref $srcref eq 'ARRAY' ? @$srcref : $srcref);
    foreach $src (@src) {
	my(@stat_orig)  = stat $src;
	if (defined $stat_orig[9] &&
	    $stat_cache[9] < $stat_orig[9]) {
	    # Cache ist nicht gltig
	    return undef;
	}
    }
    1;
}

sub get_from_cache {
    my($cachefile, $srcref) = @_;
    return if !$cacheable || !@cacheable;

    if (!defined $cachedir) {
	$cacheable = 0;
	return undef;
    }
    
    $cachefile = "$cachedir/$cachefile" unless $cachefile =~ m|^/|;
    my($cache_func_found, $cachepath) = try_cache($cachefile);
    return undef if !$cache_func_found;
    warn "OK\n" if $VERBOSE;

    return undef if !valid_cache($cachepath, $srcref);

    if ($cache_func_found eq 'Storable') {
	# Cache ist gltig
	Storable::retrieve($cachepath);
    } elsif ($cache_func_found eq 'Data::Dumper') {
	do $cachepath;
	my $x = $datadumper_var;
	undef $datadumper_var; # evtl. memory leak vermeiden
	$x;
    } elsif ($cache_func_found eq 'VirtArray') {
	my @a;
	tie @a, 'VirtArray', $cachepath;
	\@a;
    } elsif (defined $cache_func_found) {
	warn "Unknown cache function: $cache_func_found";
    }
}

sub write_cache {
    my($ref, $cachefile) = @_;
    return if !$cacheable || !@cacheable;

    $cachefile = "$cachedir/$cachefile" unless $cachefile =~ m|^/|;
    my($cache_func_found, $cachepath) = try_cache($cachefile, 1, $ref);
    return undef if !$cache_func_found;
    warn "OK\n" if $VERBOSE;

    if ($cache_func_found eq 'Storable') {
	Storable::store($ref, $cachepath);
    } elsif ($cache_func_found eq 'Data::Dumper') {
	$Data::Dumper::Indent = 1;
	if (open(DD, ">$cachepath")) {
	    binmode DD;
	    print DD Data::Dumper->Dumpxs([$ref], ['datadumper_var']);
	    close DD;
	    1;
	} else {
	    undef;
	}
    } elsif ($cache_func_found eq 'VirtArray') {
	VirtArray::store($ref, $cachepath);
    } elsif (defined $cache_func_found) {
	warn "Unknown cache function: $cache_func_found";
    }
}

######################################################################
package Strassen;
use BBBikeUtil;
#use AutoLoader 'AUTOLOAD';
use vars qw(@datadirs $OLD_AGREP $VERBOSE);

@datadirs = ("$FindBin::RealBin/data", './data');
foreach (@INC) {
    push @datadirs, "$_/data";
}

$OLD_AGREP = 0;

sub new {
    my($class, $filename, %arg) = @_;
    my(@filenames);
    if (defined $filename) {
	push(@filenames, $filename, map($_ . "/$filename", @datadirs));
    } else {
	push(@filenames, map($_ . "/strassen", @datadirs));
    }
    my $self = {};
    bless $self, $class;

  TRY: {
	my $file;
	foreach $file (@filenames) {
	    if (-f $file and -r _) {
		$self->{File} = $file;
		last TRY;
	    }
	    my $gzfile = "$file.gz";
	    if (-f $gzfile and -r _) {
		$self->{File} = $gzfile;
		$self->{IsGzipped} = 1;
	    }
	}
	die "Can't open ", join(", ", @filenames);
    }
    unless ($arg{NoRead}) {
	$self->read_data;
    }

    $self->{Pos}   = 0;
    $self->{Koord} = $arg{Koord} if exists $arg{Koord};

    $self;
}

sub read_data {
    my($self) = @_;
    my $file = $self->{File};
    if ($self->{IsGzipped}) {
	die "Can't execute zcat $file" if !open(B, "zcat $file |");
    } else {
	die "Can't open $file" if !open(B, $file);
    }
    binmode B;
    my @data;
    while (<B>) {
	next if /^\#/ || /^\s*$/;
	push @data, $_;
    }
    close(B);
    $self->{Data}  = \@data;
}

sub new_from_data {
    my($class, @data) = @_;
    my $self = {};
    $self->{Data} = \@data;
    $self->{Pos}  = 0;
    bless $self, $class;
}

sub write {
    my($self, $filename) = @_;
    if (open(COPY, ">$filename")) {
	binmode COPY;
	print COPY join("", @{$self->{Data}});
	close COPY;
	1;
    } else {
	warn "Can't write to $filename: $!" if $VERBOSE;
	0;
    }
}

sub get {
    my($self, $pos) = @_;
    my $line = $self->{Data}->[$pos];
    parse($line);
}

sub parse {
    my $line = shift;
    return [undef, [], undef] if !$line;
    my $name;
    ($name, $line) = split(/\t/, $line, 2);
    my @s = split(/\s+/, $line);
    my $category = shift @s;
    [$name, \@s, $category];
}

sub parse_rigid {
    my($name, $resref, $category) = @{parse(shift)};
    my @res;
    foreach (@$resref) {
	push(@res, $_) if $_ =~ /^\d+$/ || $_ =~ /^.?-?\d+,-?\d+$/;
    }
    [$name, \@res, $category];
}

sub get_obj {
    my($self, $pos) = @_;
    Strasse->new($self->get($pos));
}

# initialisiert fr next() und gibt *keinen* Wert zurck
sub init {
    my $self = shift;
    $self->{Pos} = -1;
}

# initialisiert fr next() und gibt den ersten Wert zurck
sub first {
    my $self = shift;
    $self->{Pos} = 0;
    $self->get(0);
}

sub next {
    my $self = shift;
    $self->get(++($self->{Pos}));
}

sub next_obj {
    my $self = shift;
    $self->get_obj(++($self->{Pos}));
}

sub at_end {
    my $self = shift;
    $self->{Pos} >= $#{$self->{Data}};
}

sub count {
    my $self = shift;
    scalar @{$self->{Data}};
}

# gibt die aktuelle Position zurck
sub pos { shift->{Pos} }

# Gibt die Positionen (als Array) fr einen bestimmten Namen zurck
# Achtung: eine laufende init/next-Schleife wird hiermit zurckgesetzt!
sub pos_from_name {
    my($self, $name) = @_;
    my @res;
    my $found = 0;
    $self->init;
    while(1) {
	my $ret = $self->next;
	last if !@{$ret->[1]};
	if ($ret->[0] eq $name) {
	    push @res, $self->pos;
	    $found++;
	} elsif ($found) {
	    last;
	}
    }
    @res;
}

# Wandelt die Indices aus dem Ergebnis von get() (2. Element) in
# Koordinaten um.
# Gibt eine Referenz auf ein Array zurck: [[x1,y1], [x2,y2] ...]
sub to_koord {
    my($self, $resref) = @_;
    my @res;
    foreach (@$resref) {
	if (/^(-?\d+),(-?\d+)$/) {
	    push(@res, [$1, $2]);
	} elsif (/(-?\d+),(-?\d+)$/) { # ignore prefix XXX
	    push(@res, [$1, $2]);
	} else {
	    warn "XXX May fail XXX";
	    push(@res, [$self->{Koord}->get($_)]);
	}
    }
    \@res;
}

# wie to_koord, nur fr einen Punkt
# XXX Koordinaten der Form prefix(x,y) bearbeiten
sub to_koord1 {
    my($self, $s) = @_;
    if ($s =~ /^(-?\d+),(-?\d+)$/) {
	[$1, $2];
    } elsif ($self->{Koord}) {
	[$self->{Koord}->get($_)];
    } elsif ($s =~ /^((:[^:]*:)?([A-Za-z])?)?(-?\d+),(-?\d+)$/) {
	# Ausgabe: x, y, coordsystem, bahnhof
	[$4, $5, $3, $2];
    } else {
	warn "No Koord object...";
	[undef, undef]; # XXX
    }
}

# Gibt alle echten Kreuzungen in einer Liste aus.
sub all_crossings {
    my($self, %args) = @_;
    my $rettype   = $args{RetType};
    my $use_cache = $args{UseCache};
    if (!defined $rettype) { $rettype = 'array' }
    my $cachefile = "all_crossings_$rettype";
    if ($use_cache && $rettype =~ /^hash/) {
	my $hashref = Strassen::Util::get_from_cache($cachefile,
						     [$self->{File}]);
	if (defined $hashref) {
	    warn "Using cache for $cachefile\n" if $VERBOSE;
	    return %$hashref;
	}
    }
    # RetType ...pos: Positionen statt Straennamen speichern
    my $store_pos = ($rettype =~ /pos$/);
    my %crossings;
    my %crossing_name;
    $self->init();
    while(1) {
	my $ret = $self->next();
	my @kreuzungen = @{$ret->[1]};
	last if @kreuzungen == 0;
	my @kreuz_coord = @{$self->to_koord(\@kreuzungen)};
	my $store = ($store_pos ? $self->pos : $ret->[0]);
	foreach (@kreuz_coord) {
	    my $xy = "$_->[0],$_->[1]";
	    $crossings{$xy}++;
	  TEST: {
		my $test;
		foreach $test (@{$crossing_name{$xy}}) {
		    last TEST if ($test eq $store);
		}
		push @{$crossing_name{$xy}}, $store;
	    }
	}
    }
    if ($rettype =~ /^hash/) { # Rckgabewert: "x,y" => [name1,name2,name3]
	my @to_del;
	while(my($k, $v) = each %crossings) {
	    if ($v < 2) {
		push @to_del, $k;
	    } else {
		$crossings{$k} = $crossing_name{$k};
	    }
	}
	foreach (@to_del) {
	    delete $crossings{$_};
	}
	if ($use_cache &&
	    Strassen::Util::write_cache(\%crossings, $cachefile)) {
	    warn "Wrote cache ($cachefile)\n" if $VERBOSE;
	}
	%crossings;
    } else {
	my @crossings;
	while(my($k, $v) = each %crossings) {
	    if ($v > 1) {
		my($x, $y) = split(/,/, $k);
		push @crossings, [$x, $y, join("/", @{$crossing_name{$k}})];
	    }
	}
	@crossings;
    }
}

sub union {
    my($self, $str_ref, %args) = @_;

    my $uniq = !$args{Nouniq};

    my %str;
    foreach (@$str_ref) {
	$str{$_->[0]}->{$_->[1]}++;
    }

    my %res;
    $self->init;
    my $last;
    while(1) {
	my $ret = $self->next;
	last if !@{$ret->[1]};
	my $name = $ret->[0];
	if ($uniq) {
	    if (defined $last && $last eq $name) {
		next;
	    } else {
		$last = $name;
	    }
	}
	my @bez;
	if ($name =~ /^(.*)\s+\((.*)\)$/) {
	    $name = $1;
	    @bez = split(/,\s*/, $2);
	}
	if (exists $str{$name}) {
	    if (@bez) {
		foreach my $bez (@bez) {
		    if (exists $str{$name}->{$bez}) {
			$res{$self->pos}++;
			last;
		    }
		}
	    } else {
		$res{$self->pos}++;
	    }
	}
    }
    keys %res;
}

sub agrep {
    my($self, $pattern, %arg) = @_;
    agrep_file($pattern, $self->{File}, %arg);
}

# XXX make gzip-aware
# %arg:
# NoDot: keine Ausgabe von "...", wenn zu viele Matches existieren
# ErrorDef: Angabe der Reihenfolge (match begin, match errors)
sub agrep_file {
    my($pattern, $file, %arg) = @_;
    my $path;
    if (-r $file) {
	$path = $file;
    } else {
	foreach (@datadirs) {
	    if (-r "$_/$file") {
		$path = "$_/$file";
		last;
	    }
	}
    }
    if (!defined $path) {
	warn "File $file not found in @datadirs.\n";
	return undef;
    }
    my $grep_type;
    my @data;
    if (!$OLD_AGREP && is_in_path('agrep')) {
	$grep_type = 'agrep';
	$pattern =~ s/(.)/\\$1/g;
    } else {
	open(F, $path) or die "Can't open $path: $!";
	chomp(@data = <F>);
	close F;
	eval { local $SIG{'__DIE__'};
	       require String::Approx;
	       String::Approx->VERSION(2.7);
	   };
	if (!$@) {
	    $grep_type = 'approx';
	} else {
	    $grep_type = 'perl';
	}
    }
    my @def;
    if ($arg{ErrorDef}) {
	@def = @{$arg{ErrorDef}};
    } else {
	@def = ([1, 0],
		[1, 1],
		[1, 2],
		[0, 0],
		[0, 1],
		[0, 2],
		[0, 3],
	       );
    }
    for my $def (@def) {
	my($begin, $err) = @$def;
	my @this_res;
	if ($grep_type eq 'agrep') {
	    my @args = '-i';
	    my $pattern = ($begin ? "^$pattern" : $pattern);
	    if ($err > 0) { push @args, "-$err" }
	    open(AGREP, "-|") or
	      exec 'agrep', @args, $pattern, $path or
		die "Can't exec program: $!";
	    chomp(@this_res = <AGREP>);
	    close AGREP;
	} elsif ($grep_type eq 'approx' && $err) {
	    next if $begin || $err > 2; # Bug bei $err == 3
	    @this_res = String::Approx::amatch($pattern, ['i', $err], @data);
	} else { # weder agrep noch String::Approx
	    my $pattern = ($begin ? "^$pattern" : $pattern);
	    if ($err == 0) { 
		@this_res = grep(/\Q$pattern\E/i, @data);
	    } elsif ($err == 1) { # metacharacter erlauben
		@this_res = grep(/$pattern/i, @data);
	    } else {
		next;
	    }
	}
	if (@this_res == 1) {
	    return parse($this_res[0])->[0];
	} elsif (@this_res) {
	    my(@res1, @res2, @res3);
	    my $i = 0;
	    my $last_name;
	    foreach (@this_res) {
		$i++;
		my $name = parse($_)->[0];
		if (defined $last_name && $last_name eq $name) {
		    next;
		} else {
		    $last_name = $name;
		}
		if ($name eq $pattern) {
		    return $name;
		} elsif ($name =~ /^\Q$pattern\E/i) {
		    push(@res1, $name);
		} elsif ($i < 20) {
		    push(@res2, $name);
		} elsif ($i == 20) {
		    push(@res3, "...") unless $arg{NoDot};
		}
	    }
	    @res1 = sort @res1;
	    @res2 = sort @res2;
	    return @res1, @res2, @res3;
	}
    }
    ();
}

# Sucht Strae anhand des Bezirkes.
sub choose_street {
    my($str, $strasse, $bezirk, %args) = @_;
    my @pos;
    $str->init;
    while(1) {
	my $ret = $str->next;
	last if !@{$ret->[1]};
	my $check_strasse = $ret->[0];
	if ($check_strasse =~ /^$strasse/) {
	    my %bez;
	    if ($check_strasse =~ /(.*)\s+\((.*)\)/) {
		$check_strasse = $1;
		foreach (split(/\s*,\s*/, $2)) {
		    $bez{$_}++;
		}
		if (exists $bez{$bezirk}) {
		    if (wantarray) {
			push @pos, $str->pos;
		    } else {
			return $str->pos;
		    }
		}
	    } elsif ($check_strasse eq $strasse) {
		if (wantarray) {
		    push @pos, $str->pos;
		} else {
		    return $str->pos;
		}
	    }
	}
    }
    if (wantarray) {
	@pos;
    } else {
	undef;
    }
}

sub copy_orig {
    my $self = shift;
    if (! -d $Strassen::Util::tmpdir) {
	warn "$Strassen::Util::tmpdir does not exist" if $VERBOSE;
	return;
    }
    my $origdir = "$Strassen::Util::tmpdir/orig";
    if (! -d $origdir) {
	mkdir $origdir, 0700;
	if (! -d $origdir) {
	    warn "Can't create $origdir: $!" if $VERBOSE;
	    return;
	}
    }
    if (!defined $self->{File}) {
	warn "File not defined" if $VERBOSE;
	return;
    }
    if (ref $self->{File} ne 'ARRAY') {
	if (!-f $self->{File}) {
	    warn "<$self->{File}> does not exist" if $VERBOSE;
	    return;
	}
    } else {
	foreach (@{$self->{File}}) {
	    if (!-f $_) {
		warn "<$_> does not exist" if $VERBOSE;
		return;
	    }
	}
    }
    require File::Basename;
    my $dest;
    if (ref $self->{File} ne 'ARRAY') {
	$dest = "$origdir/" . File::Basename::basename($self->{File});
    } else {
	$dest = "$origdir/" . File::Basename::basename($self->{File}->[0] . "-multi");
    }
    if ($self->write($dest)) {
	$self->{OrigFile} = $dest;
	1;
    } else {
	delete $self->{OrigFile};
	0;
    }
}

# Rckgabe: (Strassen-Objekt mit neuen Straen, zu lschenden Indices)
sub diff_orig {
    my $self = shift;
    require File::Basename;
    my $origdir = "$Strassen::Util::tmpdir/orig";
    my $first_file = (ref $self->{File} eq 'ARRAY'
		      ? $self->{File}->[0] . "-multi"
		      : $self->{File});
    if (!defined $self->{OrigFile}) {
	$self->{OrigFile} =
	  "$origdir/" . File::Basename::basename($first_file);
    }
    if (! -f $self->{OrigFile}) {
	warn "<$self->{OrigFile}> does not exist" if $VERBOSE;
	delete $self->{OrigFile};
	return;
    }
    if (!is_in_path("diff")) {
	warn "diff not found in path" if $VERBOSE;
	return;
    }

    my $dest = "$origdir/" . File::Basename::basename($first_file) . ".new";
    return unless $self->write($dest);

    my $curr_line = 1;
    my(@del, @add);
    open(DIFF, "diff -u $self->{OrigFile} $dest |");
    scalar <DIFF>; scalar <DIFF>; # overread header
    while(<DIFF>) {
	chomp;
	if (/^\@\@\s*-(\d+)/) {
	    $curr_line = $1;
	} elsif (/^\+(.*)/) {
	    push @add, $1;
	} elsif (/^-/) {
	    push @del, $curr_line-1; # warum -1?
	    $curr_line++;
	} elsif (!/^[ \\]/) {
	    warn "Unknown diff line: $_";
	} else {
	    $curr_line++;
	}
    }
    close DIFF;

    unlink $dest;
    my $new_s = new_from_data Strassen @add;
    ($new_s, \@del);
}

sub strip_bezirk {
    my $str = shift;
    if ($str !~ /^\s*\(/) {
	$str =~ s/\s*\(.*\)\s*$//;
    }
    $str;
}

######################################################################

package Strassen::Fast;
#use AutoLoader 'AUTOLOAD';
use vars qw(@ISA);
@ISA = qw(Strassen);

sub new {
    my($class, $filename) = @_;
    require Storable;
    my $self = retrieve $filename;
    bless $self, $class;
}

######################################################################

package Strasse;
#use AutoLoader 'AUTOLOAD';

sub new {
    my($class, $obj) = @_;
    bless $obj, $class;
}
sub name        { $_[0]->[0] }
sub coords      { $_[0]->[1] }
sub coords_list { @{$_[0]->[1]} }
sub category    { $_[0]->[2] }
sub is_empty    { @{$_[0]->[1]} == 0 }
sub coord_as_string { $_[0]->coords->[$_[1]] }
sub coord_as_list {
    my($self, $i) = @_;
    my $s = $self->coords->[$i];
    Strassen->to_koord1($s); # XXX abhngig von Strassen!
}

# gibt TRUE zurck, wenn ein Teil der Strae im angegebenen Rechteck ist
# (keine perfekte Lsung)
sub is_in {
    my($self, $x1, $y1, $x2, $y2) = @_;
    my $len = scalar $self->coords_list;
    for(my $i=0; $i<$len; $i++) {
	my($x, $y) = @{$self->coord_as_list($i)};
	if ($x >= $x1 && $x <= $x2 &&
	    $y >= $y1 && $y <= $y2) {
	    return 1;
	}
    }
    0;
}

# statische Methode
sub de_artikel {
    my($strasse) = @_;
    if ($strasse =~ /(str\.|strae\b|allee\b|chaussee\b)/i) {
	"in die";
    } elsif ($strasse =~ /(park\b|garten\b|ring\b)/i) {
	"in den";
    } elsif ($strasse =~ /(damm\b|weg\b|steig\b)/i) {
	"in den";
    } elsif ($strasse =~ /(platz\b)/i) {
	"auf den";
    } elsif ($strasse =~ /(ufer\b|gestell\b)/i) {
	"in das";
    } elsif ($strasse =~ /(brcke\b)/i) {
	"auf die";
    } elsif ($strasse =~ /(\balt\b)/i) {
	"in die Strae";
    } elsif ($strasse =~ /(\btor\b)/i) {
	"in das";
    } elsif ($strasse =~ /(\bstern\b)/i) { # mglichst am Ende
	"auf den";
    } else {
	"=>";
    }
}

######################################################################

package Kreuzungen;
#use AutoLoader 'AUTOLOAD';

sub new {
    my($class, %args) = @_;
    my $self = {};
    $self->{Hash}  = $args{Hash};
    $self->{Array} = $args{Array};
    bless $self, $class;
}

sub make_grid {
    my($self, $grid_width) = @_;
    $self->{GridWidth} = (defined $grid_width ? $grid_width : 1000);
    $self->{Grid} = {};
    while(defined(my $k = each %{$self->{Hash}})) {
	my $grid = join(",", $self->grid(split(/,/, $k)));
	push @{$self->{Grid}{$grid}}, $k;
    }
}

sub grid {
    my($self, $x, $y) = @_;
    (int($x/$self->{GridWidth}), int($y/$self->{GridWidth}));
}

sub nearest {
    my($self, $x, $y, $distance) = @_;
# XXX distance wird noch ignoriert
#    if (!defined $distance) { $distance = 1000 }
    my $xy = join(",", $x, $y);
    my($gridx, $gridy) = $self->grid($x, $y);
    my @res;
    for my $xx (-1 .. 1) {
	for my $yy (-1 .. 1) {
	    my $s = join(",", $gridx+$xx, $gridy+$yy);
	    if (defined $self->{Grid}{$s}) {
		push @res, @{$self->{Grid}{$s}};
	    }
	}
    }
    sort { Strassen::Util::strecke_s($a, $xy) 
	     <=> Strassen::Util::strecke_s($b, $xy)} @res;
}

######################################################################

package MultiBezStr;
#use AutoLoader 'AUTOLOAD';

# make gzip-aware
sub new {
    my $class = shift;
    require MyFile;
    if (MyFile::openlist(*MULTI,
			 map { "$_/multi_bez_str" } @Strassen::datadirs)) {
	my $self = {};
	while(<MULTI>) {
	    chomp;
	    my($str, $rest) = split(/\t/);
	    my(@bez) = split(/,/, $rest);
	    $self->{$str} = \@bez;
	}
	bless $self, $class;
    } else {
	undef;
    }
}

sub exists {
    my($self, $str) = @_;
    exists $self->{$str};
}

sub bezirke {
    my($self, $str) = @_;
    if ($self->exists($str)) {
	@{$self->{$str}}
    } else {
	();
    }
}

sub hash { %{$_[0]} }

######################################################################

package MultiStrassen;
#use AutoLoader 'AUTOLOAD';
use vars qw(@ISA);
@ISA = qw(Strassen);

sub new {
    my($class, @obj) = @_;

    if (@obj == 1 && $obj[0]->isa('Strassen')) {
	return $obj[0];
    }

    my $self = {};

    $self->{Data} = [];
    $self->{File} = [];
    $self->{Pos}  = 0;
    $self->{Koord} = undef; # XXX
    $self->{SourcePos} = {};

    for (@obj) {
	if (!$_->isa('Strassen')) {
	    die "Object $_ not a Strassen object";
	}
	push @{$self->{Data}}, @{$_->{Data}};
	push @{$self->{File}}, $_->{File} if defined $_->{File};
    }
    bless $self, $class;
}

######################################################################

package StrassenNetz;
#use AutoLoader 'AUTOLOAD';
use Route;
use Math::Trig qw(rad2deg);
use vars qw($VERBOSE $acos $algorithm);
BEGIN {
    eval { require POSIX };
    if ($@) {
	$acos = \&Math::Trig::acos;
    } else {
	$acos = \&POSIX::acos; # C, daher wahrscheinlich etwas schneller
    }
}

$VERBOSE   = 0;
$algorithm = 1;

sub new {
    my($class, $strassen) = @_;
    die "argument 1 is not of type Strassen"
      if !$strassen->isa('Strassen');
    my $self = {};
    $self->{Strassen} = $strassen;
    bless $self, $class;
}

# markiert die angegebenen Objekte als Quell-Objekte fr dieses StrassenNetz
sub source {
    my($self, @source) = @_;
    push(@{$self->{Source}}, @source);
}

sub is_source {
    my($self, $source) = @_;
    foreach (@{$self->{Source}}) {
	return 1 if $_ eq $source;
    }
    0;
}

# gibt die zugehrigen Quelldateien aus
sub sourcefiles {
    my $self = shift;
    if (exists $self->{Source} && @{$self->{Source}}) {
	my @src;
	foreach (@{$self->{Source}}) {
	    push @src, $_->{File};
	}
	@src;
    } else {
	$self->{Strassen}->{File};
    }
}

sub statistics {
    my $self = shift;
    my $msg = '';
    if ($self->{Strassen}) {
	$msg .= "Anzahl der Straen:    " . $self->{Strassen}->count . "\n";
    }
# XXX Ersatz finden
#    if ($self->{KoordXY}) {
#	$msg .= "Anzahl der Kreuzungen: " . scalar keys(%{$self->{KoordXY}})
#	. "\n";
#    }
    if ($self->{Net2Name}) {
	my $count = 0;
	while(my($k,$v) = each %{$self->{Net2Name}}) {
	    $count += scalar keys %$v;
	}
	$msg .= "Anzahl der Kanten:     " . $count . "\n";
    }
    $msg;
}

sub make_net_slow_1 {
    my($self, %args) = @_;
    my($cachefile);

    my $cacheable = $args{UseCache} && $Strassen::Util::cacheable;
    if ($cacheable) {
	require File::Basename;
	my(@src) = $self->sourcefiles;
	$cachefile = join("_", map { File::Basename::basename($_) } @src);
	my $net = Strassen::Util::get_from_cache("net_1_$cachefile", \@src);
	my $net2name = Strassen::Util::get_from_cache("net2name_1_$cachefile",
						      \@src);
	if (defined $net && defined $net2name) {
	    $self->{Net} = $net;
	    $self->{Net2Name} = $net2name;
	    if ($VERBOSE) {
		warn "Using cache for $cachefile\n";
	    }
	    return;
	}
    }

    if ($VERBOSE) {
	warn "Using slow version of make_net\n";
    }
    $self->{Net}      = {};        # Verbindungsnetz
    $self->{Net2Name} = {};        # Zuordnung Strecke => Straenname
    my $net      = $self->{Net};
    my $net2name = $self->{Net2Name};
    my $strassen = $self->{Strassen};
    $strassen->init();
    while(1) {
	my $ret = $strassen->next();
	my @kreuzungen = @{$ret->[1]};
	last if @kreuzungen == 0;
	my @kreuz_coord = @{$strassen->to_koord(\@kreuzungen)};
	my $i;
	for($i = 0; $i < $#kreuzungen; $i++) {
	    # Integer reicht vollkommen aus, da die Angaben sowieso in m sind
	    my $entf = int(Strassen::Util::strecke($kreuz_coord[$i],
						   $kreuz_coord[$i+1]));
 	    $net->{$kreuzungen[$i]}{$kreuzungen[$i+1]} = $entf;
 	    $net->{$kreuzungen[$i+1]}{$kreuzungen[$i]} = $entf;
 	    $net2name->{$kreuzungen[$i]}{$kreuzungen[$i+1]} = $strassen->pos;
	}
    }

    if ($cacheable) {
	Strassen::Util::write_cache($net, "net_1_$cachefile");
	Strassen::Util::write_cache($net2name, "net2name_1_$cachefile");
	if ($VERBOSE) {
	    warn "Wrote cache ($cachefile)\n";
	}
    }
}

sub make_net_slow_2 {
    my($self, %args) = @_;
    my($cachefile);

    my $cacheable = $args{UseCache} && $Strassen::Util::cacheable;
    if ($cacheable) {
	require File::Basename;
	my(@src) = $self->sourcefiles;
	$cachefile = join("_", map { File::Basename::basename($_) } @src);
	my $coord2index = Strassen::Util::get_from_cache("coord2index_2_$cachefile", \@src);
	my $index2coord = Strassen::Util::get_from_cache("index2coord_2_$cachefile", \@src);
	my $net = Strassen::Util::get_from_cache("net_2_$cachefile", \@src);
	my $index2pos = Strassen::Util::get_from_cache("index2pos_2_$cachefile",
						       \@src);
	if (defined $coord2index &&
	    defined $index2coord &&
	    defined $net         &&
	    defined $index2pos) {
	    $self->{Coord2Index} = $coord2index;
	    $self->{Index2Coord} = $index2coord;
	    $self->{Net}         = $net;
	    $self->{Index2Pos}   = $index2pos;
	    if ($VERBOSE) {
		warn "Using cache for $cachefile\n";
	    }
	    return;
	}
    }

    if ($VERBOSE) {
	warn "Using new version of make_net\n";
    }
    $self->{Net}         = [];  # Verbindungsnetz
    $self->{Index2Pos}   = [];  # Zuordnung Index-Paar => Pos im Straenfile
    $self->{Coord2Index} = {};  # Zuordnung Koordinate => Index
    $self->{Index2Coord} = [];  # Zuordnung Index => Koordinate

    my $net         = $self->{Net};
    my $index2pos   = $self->{Index2Pos};
    my $coord2index = $self->{Coord2Index};
    my $index2coord = $self->{Index2Coord};
    my $strassen    = $self->{Strassen};
    $strassen->init();
    my $pos = 0;
local($^W)=0; #XXX
    while(1) {
	my $ret = $strassen->next();
	my @kreuzungen = @{$ret->[1]};
	last if @kreuzungen == 0;

	my @kreuz_coord = @{$strassen->to_koord(\@kreuzungen)};
	my @k_i;
	foreach my $cp (@kreuz_coord) {
	    my $c = pack("l2", @$cp);
	    if (!exists $coord2index->{$c}) {
		$coord2index->{$c} = pack("l", $pos);
		$index2coord->[$pos] = $c;
		$pos++;
	    }
	    push @k_i, $coord2index->{$c};
	}
      
	for (my $i; $i < $#k_i; $i++) {
	    my $entf = pack("l",
			    int(Strassen::Util::strecke($kreuz_coord[$i],
							$kreuz_coord[$i+1])));
	    my $k_i_u  = unpack("l", $k_i[$i]);
	    my $k_i1_u = unpack("l", $k_i[$i+1]);
	    $net->[$k_i_u]  .= $k_i[$i+1] . $entf;
	    $net->[$k_i1_u] .= $k_i[$i]   . $entf;
	    $index2pos->[$k_i_u]  .= $k_i[$i+1] . pack("l", $strassen->pos);
	    $index2pos->[$k_i1_u] .= $k_i[$i]   . pack("l", $strassen->pos);
	}
    }

    if ($cacheable) {
	Strassen::Util::write_cache($coord2index, "coord2index_2_$cachefile");
	Strassen::Util::write_cache($index2coord, "index2coord_2_$cachefile");
	Strassen::Util::write_cache($net, "net_2_$cachefile");
	Strassen::Util::write_cache($index2pos, "index2pos_2_$cachefile");
	if ($VERBOSE) {
	    warn "Wrote cache ($cachefile)\n";
	}
    }
}

*make_net = \&make_net_slow_1;

sub make_sperre_1 {
    my($self, $sperre_file, %args) = @_;

    my %sperre_type;
    if (exists $args{Type}) {
	foreach (@{$args{Type}}) {
	    if ($_ eq 'einbahn') {
		$sperre_type{1} = 1;
	    } elsif ($_ eq 'sperre') {
		$sperre_type{2} = 1;
	    } elsif ($_ eq 'tragen') {
		$sperre_type{0} = 1;
	    } else {
		die "Unknown sperre_type: $_";
	    }
	}
    } else {
	%sperre_type = (2 => 1, 1 => 1); # Standard: einbahn und sperre
    }

    my $sperre_obj = new Strassen $sperre_file;
    my $net      = $self->{Net};
#    my $koordxy  = $self->{KoordXY};
    $sperre_obj->init;
    while(1) {
	my $ret = $sperre_obj->next;
	last if !@{$ret->[1]};
	my $category = $ret->[2];
	next if !exists $sperre_type{$category};
	my @kreuzungen = @{$ret->[1]};
	if (@kreuzungen == 1) {
	    foreach (keys %{$net->{$kreuzungen[0]}}) {
		delete $net->{$kreuzungen[0]}{$_};
		delete $net->{$_}{$kreuzungen[0]};
#		delete $koordxy->{$kreuzungen[0]};
	    }
	} else {
	    my $i;
	    for($i = 0; $i < $#kreuzungen; $i++) {
		delete $net->{$kreuzungen[$i]}{$kreuzungen[$i+1]};
		if ($category =~ /^2/) {
		    delete $net->{$kreuzungen[$i+1]}{$kreuzungen[$i]};
		}
		# XXX delete $koordxy ?
	    }
	}
    }
}

*make_sperre = \&make_sperre_1;

sub reset {
    my $self = shift;
    $self->del_add_net;
}

sub search_all {
    my($self, $from, $to) = @_;

    if (!$self->reachable($from) or
	!$self->reachable($to)) {
	return undef;
    }

    # Ein Element in @all_paths ist eine Referenz auf ein Array mit
    # folgenden Elementen:
    # 0: Liste der Pfad-Koordinaten
    # 1: Gesamtlnge des Pfads
    my @all_paths = 
      (
       [[$from], 0],
      );
    my(@found_paths, @suspended_paths);
    my %visited = ( $from => 0 );
    while (1) {
	while (@all_paths) {
	    my @new_all_paths;
	    foreach my $path_def (@all_paths) {
		my @path = @{$path_def->[0]};
		my $curr_len = $path_def->[1];
		my $last_node = $path[$#path];
		while(my($next_node, $len)
		      = each %{$self->{Net}{$last_node}}) {
		    my $next_node_len = $len + $curr_len;
		    if (defined $visited{$next_node} and
			$next_node_len >= $visited{$next_node}) {
			next;
		    }
		    $visited{$next_node} = $next_node_len;
		    if ($next_node eq $to) {
			my @koords = (@path, $to);
			warn "Found path, len: $next_node_len\n"
			    if $VERBOSE;
			push(@found_paths,
			     [$self->{Strassen}->to_koord(\@koords),
			      $next_node_len]
			    );
			next;
		    }
#		    my $virt_len = $next_node_len  + Strassen::Util::strecke
#		      ($self->{KoordXY}{$next_node}, $self->{KoordXY}{$to});
		    my $virt_len = $next_node_len
		      + Strassen::Util::strecke_s($next_node, $to);
		    if (defined $visited{$to} and $visited{$to} < $virt_len) {
			# kann nicht mehr in einer besseren Strecke
			# erreicht werden
			next;
		    }
		    push(@new_all_paths, [[@path, $next_node],
					  $next_node_len,
					  $virt_len]);
		}
	    }
	    @all_paths = sort { $a->[2] <=> $b->[2] } 
	                 (@new_all_paths, @suspended_paths);
            @suspended_paths = splice(@all_paths, 5);
	}
	if (@suspended_paths) {
	    @all_paths = sort { $a->[2] <=> $b->[2] } splice
	      (@suspended_paths, 0, 5);
	} else {
	    last;
	}
    }

    if (!@found_paths) {
	warn "Nothing found!\n" if $VERBOSE;
	();
    } else {
	@found_paths = sort { $a->[1] <=> $b->[1] } @found_paths;
	@found_paths;
    }
}

# Suche mit Ampeln
sub search_all_extended {
    my($self, $from, $to, %args) = @_;

    if (!$self->reachable($from) or
	!$self->reachable($to)) {
	return undef;
    }

    my $ampel_penalty = $args{AmpelnPenalty} || 100; # 100m Aufschlag pro Ampel
    my %ampeln = %{$args{Ampeln}};

    my @all_paths = 
      (
       [[$from], 0, undef, 0, 0],
      );
    my(@found_paths, @suspended_paths);
    my %visited = ( $from => 0 );
    while (1) {
	while (@all_paths) {
	    my @new_all_paths;
	    foreach my $path_def (@all_paths) {
		# Format von $path_def:
		# 0: Referenz auf Pfad
		# 1: derzeitige Lnge
		# 2: virtuelle Penalty
		# 3: wahre Penalty
		# 4: Anzahl der Ampeln auf der Strecke
		my @path = @{$path_def->[0]};
		my $curr_len = $path_def->[1];
		my $curr_pen = $path_def->[3];
		my $ampeln   = $path_def->[4];
		my $last_node = $path[$#path];
		while(my($next_node, $len)
		      = each %{$self->{Net}{$last_node}}) {
		    my $next_node_pen = $len + $curr_pen;
		    if (defined $visited{$next_node} and
			$next_node_pen >= $visited{$next_node}) {
			next;
		    }
		    # XXX Ampel-Code nach der berprfung auf Ziel
		    # bearbeiten
		    my($x,$y) = @{Strassen->to_koord1($next_node)};
		    if (exists $ampeln{$x}->{$y}) {
			$next_node_pen += $ampel_penalty;
			if (defined $visited{$next_node} and
			    $next_node_pen >= $visited{$next_node}) {
			    next;
			}
			$ampeln++;
		    }
		    $visited{$next_node} = $next_node_pen;
		    my $next_node_len = $len + $curr_len;
		    if ($next_node eq $to) {
			my @koords = (@path, $to);
			warn "Found path, len: $next_node_len\n"
			    if $VERBOSE;
			push(@found_paths,
			     [$self->{Strassen}->to_koord(\@koords),
			      $next_node_len,
			      undef,
			      $next_node_pen,
			      $ampeln,
			     ]
			    );
			next;
		    }
#		    my $virt_pen = $next_node_pen  + Strassen::Util::strecke
#		      ($self->{KoordXY}{$next_node}, $self->{KoordXY}{$to});
		    my $virt_pen = $next_node_pen 
		      + Strassen::Util::strecke_s($next_node, $to);
		    if (defined $visited{$to} and $visited{$to} < $virt_pen) {
			# kann nicht mehr in einer besseren Strecke
			# erreicht werden
			next;
		    }
		    push(@new_all_paths, [[@path, $next_node],
					  $next_node_len,
					  $virt_pen,
					  $next_node_pen,
					  $ampeln,
					 ]);
		}
	    }
	    @all_paths = sort { $a->[2] <=> $b->[2] } 
	                 (@new_all_paths, @suspended_paths);
            @suspended_paths = splice(@all_paths, 5);
	}
	if (@suspended_paths) {
	    @all_paths = sort { $a->[2] <=> $b->[2] } splice
	      (@suspended_paths, 0, 5);
	} else {
	    last;
	}
    }

    if (!@found_paths) {
	warn "Nothing found!\n" if $VERBOSE;
	();
    } else {
	@found_paths = sort { $a->[3] <=> $b->[3] } @found_paths;
	@found_paths;
    }
}

# mgliche Rckgabewerte:
# - die beste Pfadbeschreibung (+ Lnge etc.)
# - die besten Pfadbeschreibungen (+ Lnge etc.)
# - die beste Pfadbeschreibung (ohne Lnge etc.)
# - die besten Pfadbeschreibungen (ohne Lnge etc.)
# - die beste Route (als Objekt)
# - die besten Routen (als Objekt)
sub build_search_code {
    my($self, %args) = @_;

    my $has_penalty = (exists $args{Ampeln} || exists $args{Strassen});
    my $has_ampeln  = exists $args{Ampeln};
    my $ampel_penalty;
    if ($has_ampeln) {
	$ampel_penalty = $args{Ampeln}->{Penalty} || 100;
    }
    my $use_2 = 0;
    if (defined $args{Use2}) {
	$use_2 = $args{Use2};
    } elsif ($algorithm == 2) {
	$use_2 = 1;
    }

    my $code = 'sub {
    my($self, $from, $to) = @_;
';

    if ($use_2) {
	$code .= '
    $from = unpack("l", $self->{Coord2Index}{pack("l2", split(/,/, $from))});
    $to   = unpack("l", $self->{Coord2Index}{pack("l2", split(/,/, $to))});
';
    }

    # Format von $path_def:
    # 0: Referenz auf Pfad
    # 1: derzeitige Lnge
    # 2: virtuelle Penalty (ab hier nur wenn $has_penalty wahr ist)
    # 3: wahre Penalty
    # 4: Anzahl der Ampeln auf der Strecke (falls Ampeln eingestellt wurden)

    $code .= '
    my @all_paths = 
      (
       [[$from], 0';
    if ($has_penalty) {
	$code .= ', undef, 0';
	if ($has_ampeln) {
	    $code .= ', 0';
	}
    }
    $code .= '],
      );
    my(@found_paths, @suspended_paths);
    my %visited = ( $from => 0 );
    while (1) {
	while (@all_paths) {
	    my @new_all_paths;
	    foreach my $path_def (@all_paths) {
		my @path = @{$path_def->[0]};
		my $curr_len = $path_def->[1];
';
    if ($has_penalty) {
	$code .= '
		my $curr_pen = $path_def->[3];
';
	if ($has_ampeln) {
	    $code .= '
		my $ampeln   = $path_def->[4];
';
	}
    }
    $code .= '
		my $last_node = $path[$#path];
';
    if ($use_2) {
	$code .= '
                my $net_s = $self->{Net}[$last_node];
                my $net_s_len = length($net_s);
                for(my $i = 0; $i < $net_s_len; $i+=8) {
                    my $next_node = unpack("l", substr($net_s, $i, 4));
                    my $len = unpack("l", substr($net_s, $i+4, 4));
';
    } else {
	$code .= '
		while(my($next_node, $len)
		      = each %{$self->{Net}{$last_node}}) {
';
    }
    my $len_pen    = ($has_penalty ? 'pen' : 'len');
    my $sort_index = ($has_penalty ? 3 : 1);
    $code .= '
		    my $next_node_'.$len_pen.' = $len + $curr_'.$len_pen . ';
		    if (defined $visited{$next_node} and
			$next_node_'.$len_pen.' >= $visited{$next_node}) {
			next;
		    }
';
    if ($has_ampeln) {
	$code .= '
		    my($x,$y) = @{Strassen->to_koord1($next_node)};
		    if (exists $ampel_net->{$x}{$y}) {
			$next_node_pen += ' . $ampel_penalty . ';
			if (defined $visited{$next_node} and
			    $next_node_pen >= $visited{$next_node}) {
			    next;
			}
			$ampeln++;
		    }
';
    }
    $code .= '
		    $visited{$next_node} = $next_node_'.$len_pen.';
';
    if ($has_penalty) {
	$code .= '
		    my $next_node_len = $len + $curr_len;
';
    }
    $code .= '
		    if ($next_node eq $to) {
			my @koords = (@path, $to);
';
    if ($use_2) {
	$code .= '
	                foreach (@koords) {
                            $_ = join(",", unpack("l2", $self->{Index2Coord}[$_]));
			}
';
    }
    $code .= '
			warn "Found path, len: $next_node_len\n"
			    if $VERBOSE;
			push(@found_paths,
			     [$self->{Strassen}->to_koord(\@koords),
			      $next_node_len';
    if ($has_penalty) {
	$code .= ',
			      undef,
			      $next_node_pen,
';
	if ($has_ampeln) {
	    $code .= '			      $ampeln,
';
	}
    }
    $code .= '			     ]
			    );
			next;
		    }
		    my $virt_'.$len_pen.' = $next_node_'.$len_pen;
    if ($use_2) {
	$code .= '
		      + Strassen::Util::strecke_i($self, $next_node, $to);
';
    } else {
	$code .= '
		      + Strassen::Util::strecke_s($next_node, $to);
';
    }
    $code .= '
		    if (defined $visited{$to} and $visited{$to} < $virt_'.$len_pen.') {
			# kann nicht mehr in einer besseren Strecke
			# erreicht werden
			next;
		    }
		    push(@new_all_paths, [[@path, $next_node],
					  $next_node_len,
					  $virt_'.$len_pen.',
';
    if ($has_penalty) {
	$code .= '					  $next_node_pen,
';
	if ($has_ampeln) {
	    $code .= '					  $ampeln,';
	}
    }
    $code .= '					 ]);
		}
	    }
	    @all_paths = sort { $a->[2] <=> $b->[2] } 
	                 (@new_all_paths, @suspended_paths);
            @suspended_paths = splice(@all_paths, 10); # XXX 10 berprfen
	}
	if (@suspended_paths) {
	    @all_paths = sort { $a->[2] <=> $b->[2] } splice
	      (@suspended_paths, 0, 5);
	} else {
	    last;
	}
    }

    if (!@found_paths) {
	warn "Nothing found!\n" if $VERBOSE;
        ';
    $code .= ($args{All} ? '()' : 'undef');
    $code .= ';
    } else {
	@found_paths = sort { $a->['.$sort_index.'] <=> $b->['.$sort_index.'] } @found_paths;
        ';
    $code .= ($args{All} 
	      ? '@found_paths' 
	      : ($args{OnlyPath}
		 ? '$found_paths[0]->[0]'
		 : '$found_paths[0]'
		)
	     );
    $code .= ';
    }
}
';

    $code;
}

sub new_search {
    my($self, $from, $to, %args) = @_;

    my $ampel_net;
    if (exists $args{Ampeln}) {
	$ampel_net = $args{Ampeln}->{Net};
    }

    $args{OnlyPath} = 1;
    my $code = $self->build_search_code(%args);
    if ($VERBOSE) {
	my $i = 0;
	foreach (split(/\n/, $code)) {
	    $i++;
	    printf STDERR "%3d %s\n", $i, $_;
	}
    }
    my $search_sub = eval $code;

    if (exists $args{Via} and @{$args{Via}}) {
	my(@route) = ($from, @{$args{Via}}, $to);
	my @path;
	my $i;
	for($i = 0; $i < $#route; $i++) {
	    my $search_res = $search_sub->($self, $route[$i], $route[$i+1]);
	    if (ref $search_res eq 'ARRAY') {
		my(@found_path) = @$search_res;
		if ($i > 0) {
		    shift @found_path;
		}
		push(@path, @found_path);
	    }
	}
	\@path;
    } else {
	$search_sub->($self, $from, $to);
    }
}

sub search_all_obj {
    my($self, $from, $to, %args) = @_;
    my(@found_paths);
    if (keys %args) {
	@found_paths = $self->search_all_extended($from, $to, %args);
    } else {
	@found_paths = $self->search_all($from, $to);
    }
    my @res;
    foreach (@found_paths) {
	push(@res, Route->new(Path   => $_->[0],
			      Len    => $_->[1],
			      From   => $from,
			      To     => $to,
			      Ampeln => $_->[4]));
    }
    @res;
}

sub search {
    my($self, $from, $to, %args) = @_;
    my(@found_paths);
    if (keys %args) {
	@found_paths = $self->search_all_extended($from, $to, %args);
    } else {
	@found_paths = $self->search_all($from, $to);
    }
    if (@found_paths) {
	$found_paths[0]->[0]; # krzeste Strecke
    } else {
	[];
    }
}

sub search_obj {
    my($self, $from, $to, %args) = @_;
    my(@found_paths);
    if (keys %args) {
	@found_paths = $self->search_all_extended($from, $to, %args);
    } else {
	@found_paths = $self->search_all($from, $to);
    }
    if (@found_paths) {
	Route->new(Path   => $found_paths[0]->[0],
		   Len    => $found_paths[0]->[1],
		   From   => $from,
		   To     => $to,
		   Ampeln => $found_paths[0]->[4]);
    } else {
	undef;
    }
}

sub search_via {
    my($self, $from, $to, $via_ref, %args) = @_;
    my(@route) = ($from, @$via_ref, $to);
    my @path;
    my $i;
    for($i = 0; $i < $#route; $i++) {
	my $search_res = $self->search($route[$i], $route[$i+1], %args);
	if (ref $search_res eq 'ARRAY') {
	    my(@found_path) = @$search_res;
	    if ($i > 0) {
		shift @found_path;
	    }
	    push(@path, @found_path);
	}
    }
    \@path;
}

sub search_via_obj {
    my($self, $from, $to, $via_ref, %args) = @_;
    my(@route) = ($from, @$via_ref, $to);
    my @r;
    my $i;
    for($i = 0; $i < $#route; $i++) {
	push(@r, $self->search_obj($route[$i], $route[$i+1], %args));
    }
    Route::concat(@r);
}

sub search_via_obj_all {
    my($self, $from, $to, $via_ref, %args) = @_;
    if (!@$via_ref) {
	$self->search_all_obj($from, $to, %args);
    } else {
	$self->search_via_obj($from, $to, $via_ref, %args);
    }
}

# Erzeugt aus einer Route eine Liste mit den zugehrigen Straennamen.
# Die Elemente der Liste bestehen aus:
#    [Straenname, Entfernung (m), Winkel]
# XXX
sub route_to_name_1 {
    my($self, $route_ref) = @_;
    my @strname;
    my $i;
    # XXX kann beim acos anscheinend auftreten
    local $SIG{FPE} = sub { warn "Caught SIGFPE!" };
    for($i = 0; $i < $#{$route_ref}; $i++) {
	my $xy1 = Route::_coord_as_string([$route_ref->[$i][0],
					   $route_ref->[$i][1]]);
	my $xy2 = Route::_coord_as_string([$route_ref->[$i+1][0],
					   $route_ref->[$i+1][1]]);
	my $str_i;
	if (exists $self->{Net2Name}{$xy1}{$xy2}) {
	    $str_i = $self->{Net2Name}{$xy1}{$xy2};
	} elsif (exists $self->{Net2Name}{$xy2}{$xy1}) {
	    $str_i = $self->{Net2Name}{$xy2}{$xy1};
	} else {
	    $str_i = undef;
	}
	my $str = (defined $str_i
		   ? ($str_i =~ /^\d/
		      ? $self->{Strassen}->get($str_i)->[0] 
		      : $str_i
		     )
		   : "...");
	my $entf = $self->{Net}{$xy1}{$xy2};
	if (!defined $entf) {
	    $entf = Strassen::Util::strecke($route_ref->[$i],
					    $route_ref->[$i+1]);
	}
	my($winkel, $richtung);
	if ($i+1 < $#{$route_ref}) {
	    my $a_len = Strassen::Util::strecke($route_ref->[$i],
						$route_ref->[$i+1]);
	    my $b_len = Strassen::Util::strecke($route_ref->[$i+1],
						$route_ref->[$i+2]);
	    my $a1 = $route_ref->[$i+1][0]-$route_ref->[$i][0];
	    my $a2 = $route_ref->[$i+1][1]-$route_ref->[$i][1];
	    my $b1 = $route_ref->[$i+2][0]-$route_ref->[$i+1][0];
	    my $b2 = $route_ref->[$i+2][1]-$route_ref->[$i+1][1];
	    $richtung = ($a1*$b2-$a2*$b1 > 0 ? 'l' : 'r');
	    $winkel = ($a_len == 0 || $b_len == 0 ? 0 
		       : rad2deg(&$acos(($a1*$b1+$a2*$b2)/($a_len*$b_len))));
	    $winkel = -$winkel if $winkel < 0; # bei altem Math::Trig::acos 
	}
	if (@strname && $str eq $strname[$#strname]->[0]) {
	    $strname[$#strname]->[1] += $entf;
	    $strname[$#strname]->[2] = $winkel;
	    $strname[$#strname]->[3] = $richtung;
	} else {
	    my @val = ($str, $entf, $winkel, $richtung);
	    push @strname, [@val];
	}

    }

    @strname;
}

sub route_to_name_2 {
    my($self, $route_ref) = @_;
    my @strname;
    my $i;
    # XXX kann beim acos anscheinend auftreten
    local $SIG{FPE} = sub { warn "Caught SIGFPE!" };
    for($i = 0; $i < $#{$route_ref}; $i++) {
	my $xy1 = $self->{Coord2Index}->
	  {pack("l2", $route_ref->[$i][0], $route_ref->[$i][1])};
	my $xy1_u = unpack("l", $xy1);
	my $xy2 = $self->{Coord2Index}->
	  {pack("l2", $route_ref->[$i+1][0], $route_ref->[$i+1][1])};
	my $str_i;
	{
	    my $net_s = $self->{Index2Pos}[$xy1_u];
	    my $net_s_len = length($net_s);
	    for(my $i = 0; $i < $net_s_len; $i+=8) {
		if (substr($net_s, $i, 4) eq $xy2) {
		    $str_i = unpack("l", substr($net_s, $i+4, 4));
		    last;
		}
	    }
	}
# 	if (exists $self->{Index2Pos}{$xy1}{$xy2}) {
# 	    $str_i = $self->{Index2Pos}{$xy1}{$xy2};
# 	} elsif (exists $self->{Index2Pos}{$xy2}{$xy1}) {
# 	    $str_i = $self->{Index2Pos}{$xy2}{$xy1};
# 	} else {
# 	    $str_i = undef;
# 	}
	my $str = (defined $str_i
		   ? ($str_i =~ /^\d/
		      ? $self->{Strassen}->get($str_i)->[0] 
		      : $str_i
		     )
		   : "...");
	my $entf;
	{
	    my $net_s = $self->{Net}[$xy1_u];
	    my $net_s_len = length($net_s);
	    for(my $i = 0; $i < $net_s_len; $i+=8) {
		if (substr($net_s, $i, 4) eq $xy2) {
		    $entf = unpack("l", substr($net_s, $i+4, 4));
		    last;
		}
	    }
	}
	if (!defined $entf) {
	    $entf = Strassen::Util::strecke($route_ref->[$i],
					    $route_ref->[$i+1]);
	}
	my($winkel, $richtung);
	if ($i+1 < $#{$route_ref}) {
	    my $a_len = Strassen::Util::strecke($route_ref->[$i],
						$route_ref->[$i+1]);
	    my $b_len = Strassen::Util::strecke($route_ref->[$i+1],
						$route_ref->[$i+2]);
	    my $a1 = $route_ref->[$i+1][0]-$route_ref->[$i][0];
	    my $a2 = $route_ref->[$i+1][1]-$route_ref->[$i][1];
	    my $b1 = $route_ref->[$i+2][0]-$route_ref->[$i+1][0];
	    my $b2 = $route_ref->[$i+2][1]-$route_ref->[$i+1][1];
	    $richtung = ($a1*$b2-$a2*$b1 > 0 ? 'l' : 'r');
	    $winkel = ($a_len == 0 || $b_len == 0 ? 0 
		       : rad2deg(&$acos(($a1*$b1+$a2*$b2)/($a_len*$b_len))));
	    $winkel = -$winkel if $winkel < 0; # bei altem Math::Trig::acos 
	}
	if (@strname && $str eq $strname[$#strname]->[0]) {
	    $strname[$#strname]->[1] += $entf;
	    $strname[$#strname]->[2] = $winkel;
	    $strname[$#strname]->[3] = $richtung;
	} else {
	    push @strname, [$str, $entf, $winkel, $richtung];
	}

    }

    @strname;
}

*route_to_name = \&route_to_name_1;

sub add_net {
    my($self, $pos, @points) = @_;
    die 'Es mssen genau 3 Punkte in @points sein!' if @points != 3;
    my($startx, $starty) = @{$points[0]};
    my $starts = Route::_coord_as_string([$startx,$starty]);
    foreach (@points) {
	my($x, $y) = @$_;
	my $s = Route::_coord_as_string([$x,$y]);
	next if $s eq $starts;
	my $entf = Strassen::Util::strecke($points[0], $_);
	if (!exists $self->{Net}{$starts}{$s}) {
	    $self->{Net}{$starts}{$s} = $entf;
	    push(@{$self->{AdditionalNet}}, [$starts, $s]);
	}
	if (!exists $self->{Net}{$s}{$starts}) {
	    $self->{Net}{$s}{$starts} = $entf;
	    push(@{$self->{AdditionalNet}}, [$s, $starts]);
	}
	if (!exists $self->{Net2Name}{$starts}{$s} &&
	    defined $pos) {
	    $self->{Net2Name}{$starts}{$s} = $pos;
	}
    }
}

sub add_faehre {
    my($self, $faehre_file, %args) = @_;
    my $faehre_obj = new Strassen $faehre_file;
    $faehre_obj->init;
    while(1) {
	my $ret = $faehre_obj->next;
	last if !@{$ret->[1]};
	my @kreuzungen = @{$ret->[1]};
	my $i;
	# XXX record to make deletion possible
	for($i = 1; $i<=$#kreuzungen; $i++) {
	    $self->{Net}{$kreuzungen[$i-1]}{$kreuzungen[$i]} = 0;
	    $self->{Net}{$kreuzungen[$i]}{$kreuzungen[$i-1]} = 0;
	    $self->{Net2Name}{$kreuzungen[$i-1]}{$kreuzungen[$i]} =
	      "Fhre " . $ret->[0];
	}
    }
}

sub del_add_net {
    my $self = shift;
    my $a;
#    foreach $a (@{$self->{Additional}}) {
#	delete $self->{KoordXY}{$a};
#    }
    my $b;
    foreach $b (@{$self->{AdditionalNet}}) {
	delete $self->{Net}{$b->[0]}{$b->[1]};
	if (exists $self->{Net2Name}{$b->[0]}{$b->[1]}) {
	    delete $self->{Net2Name}{$b->[0]}{$b->[1]};
	}
    }
    @{$self->{Additional}} = ();
    @{$self->{AdditionalNet}} = ();
}	  

sub reachable_1 {
    my($self, $index) = @_;
    if (!exists $self->{Net}{$index}) {
	warn "$index is not reachable\n" if $VERBOSE;
	0;
    } else {
	1;
    }
}

sub reachable_2 {
    my($self, $coord) = @_;
    if (!defined $self->{Net}[$self->{Coord2Index}{$coord}]) {
	warn "$coord is not reachable\n" if $VERBOSE;
	0;
    } else {
	1;
    }
}

*reachable = \&reachable_1;

sub null { }

sub use_algorithm {
    my $a = shift;
    if (defined $a) {
	$algorithm = $a;
    }

    local($^W) = 0;

    *make_net      = ($algorithm == 1 ? \&make_net_slow_1 : \&make_net_slow_2);
    *make_sperre   = ($algorithm == 1 ? \&make_sperre_1 : \&null);
    *route_to_name = ($algorithm == 1 ? \&route_to_name_1 : \&route_to_name_2);
    *reachable     = ($algorithm == 1 ? \&reachable_1 : \&reachable_2);
    # XXX restliche ...
}

1;
