package VirtArray;

use strict;
use Carp;
use vars qw($VERSION @ISA $magic $VERBOSE);

require DynaLoader;

@ISA = qw(DynaLoader);
$VERSION = '0.01';

$magic = "FARR";
$VERBOSE = 0;

bootstrap VirtArray $VERSION;

sub retrieve {
    my($file) = @_;
    my $buf;
    my @res;
    open(F, $file) or croak "Can't open $file: $!";
    read F, $buf, 4;
    croak "Wrong magic" if ($buf ne $magic);
    my $is_var_len;
    read F, $buf, 4;
    $is_var_len = unpack("l", $buf);
    my $len;
    read F, $buf, 4;
    $len = unpack("l", $buf);
    if (!$is_var_len) {
	read F, $buf, 4;
	my $reclen = unpack("l", $buf);
	print STDERR "Reading fixed data (len=$reclen, rec=$len)\n"
	  if $VERBOSE;
	for(my $i=0; $i<$len; $i++) {
	    read F, $buf, $reclen;
	    push @res, $buf;
	}
    } else {
	print STDERR "Reading variable data (rec=$len)\n" if $VERBOSE;
	my @index;
	for(my $i=0; $i<=$len; $i++) {
	    read F, $buf, 4;
	    push @index, unpack("l", $buf);
	}
	for(my $i=0; $i<$len; $i++) {
	    my $reclen = $index[$i+1]-$index[$i];
	    read F, $buf, $reclen;
	    push @res, $buf;
	}
    }
    close F;
    \@res;
}

sub store {
    my($array_ref, $file) = @_;
    croak "Not an array reference" if (!ref $array_ref eq 'ARRAY');
    croak "Empty array" if (@$array_ref == 0);
    $file = "-" if !defined $file;
    my $is_var_len = 0;
    my $reclen = length($array_ref->[0]);
    for(my $i=1; $i<=$#$array_ref; $i++) {
	if (length($array_ref->[$i]) != $reclen) {
	    $is_var_len = 1;
	    last;
	}
    }
    # Format:
    # magic
    # long (0: fixed length, 1: var length)
    # long (number of records)
    open(F, ">$file") or croak "Can't open $file: $!";
    print STDERR "Writing to $file\n" if $VERBOSE;
    print F $magic;
    print F pack("l", $is_var_len);
    my $len = scalar @$array_ref;
    print F pack("l", $len);
    if (!$is_var_len) {
	print STDERR "Writing fixed data (len=$reclen, rec=$len)\n"
	  if $VERBOSE;
	# Format:
	# long (length of record)
	# data
	print F pack("l", $reclen);
	foreach (@$array_ref) {
	    print F $_;
	}
    } else {
	print STDERR "Writing variable data (rec=$len)\n" if $VERBOSE;
	my $data = '';
	my $i = 0; # $i == length($data)
	foreach (@$array_ref) {
	    $data .= $_;
	    print F pack("l", $i);
	    $i += length($_);
	}
	print F pack("l", $i); # letzter Index
	print F $data;
    }
    close F;
}

sub is_valid {
    my $filename = shift;

    if (open(F, $filename)) {
	my $read_magic;
	read(F, $read_magic, 4);
	close F;
	return $magic eq $read_magic;
    }

    undef;
}

1;
__END__
# Below is the stub of documentation for your module. You better edit it!

=head1 NAME

VirtArray - Perl extension for blah blah blah

=head1 SYNOPSIS

  use VirtArray;
  blah blah blah

=head1 DESCRIPTION

Stub documentation for VirtArray was created by h2xs. It looks like the
author of the extension was negligent enough to leave the stub
unedited.

Blah blah blah.

=head1 AUTHOR

A. U. Thor, a.u.thor@a.galaxy.far.far.away

=head1 SEE ALSO

perl(1).

=cut
