# -*- perl -*-

#
# $Id: Enscript.pm,v 1.1 1998/05/17 17:50:31 eserte Exp eserte $
# Author: Slaven Rezic
#
# Copyright (C) 1998 Slaven Rezic. All rights reserved.
# This program is free software; you can redistribute it and/or
# modify it under the same terms as Perl itself.
#
# Mail: eserte@cs.tu-berlin.de
# WWW:  http://user.cs.tu-berlin.de/~eserte/
#

package Tk::Enscript;
use Tk;
use Tk::Font;
use Text::Tabs;
use Exporter;

use strict;
use vars qw(%media %postscript_to_x11_font
	    $VERSION @ISA @EXPORT);

@ISA = qw(Exporter);
@EXPORT = qw(enscript);
$VERSION = '0.01';

parse_cfg();

# %postscript_to_x11_font =
#   (qw(
#       bookman-demi
# 			-*-bookman-bold-r-normal--*-%d-*-*-*-*-iso8859-1
#       bookman-demiitalic
# 			-*-bookman-bold-i-normal--*-%d-*-*-*-*-iso8859-1
#       bookman-light
# 			-*-bookman-medium-r-normal--*-%d-*-*-*-*-iso8859-1
#       bookman-lightitalic
# 			-*-bookman-medium-i-normal--*-%d-*-*-*-*-iso8859-1
#       courier-bold
# 			-*-courier-bold-o-normal--*-%d-*-*-*-*-iso8859-1
#       courier-boldoblique
# 			-*-courier-bold-r-normal--*-%d-*-*-*-*-iso8859-1
#       courier-oblique
# 			-*-courier-medium-o-normal--*-%d-*-*-*-*-iso8859-1
#       courier
# 			-*-courier-medium-r-normal--*-%d-*-*-*-*-iso8859-1
#       helvetica-bold
# 			-*-helvetica-bold-r-normal--*-%d-*-*-*-*-iso8859-1
#       helvetica-boldoblique
# 			-*-helvetica-bold-o-normal--*-%d-*-*-*-*-iso8859-1
#       helvetica-oblique
# 			-*-helvetica-medium-o-normal--*-%d-*-*-*-*-iso8859-1
#       helvetica
# 			-*-helvetica-medium-r-normal--*-%d-*-*-*-*-iso8859-1
#       times-bold
# 			-*-times-bold-r-normal--*-%d-*-*-*-*-iso8859-1
#       times-bolditalic
# 			-*-times-bold-i-normal--*-%d-*-*-*-*-iso8859-1
#       times-italic
# 			-*-times-medium-i-normal--*-%d-*-*-*-*-iso8859-1
#       times-roman
# 			-*-times-medium-r-normal--*-%d-*-*-*-*-iso8859-1
#      ),
#    'avantgarde-book' => undef,
#    'avantgarde-bookoblique' => undef,
#    'avantgarde-demi' => undef,
#    'avantgarde-demioblique' => undef,
#    'helvetica-narrow-bold' => undef,
#    'helvetica-narrow-boldoblique' => undef,
#    'helvetica-narrow-oblique' => undef,
#    'helvetica-narrow' => undef,
#    'newcenturyschlbk-bold' => undef,
#    'newcenturyschlbk-bolditalic' => undef,
#    'newcenturyschlbk-italic' => undef,
#    'newcenturyschlbk-roman' => undef,
#    'palatino-bold' => undef,
#    'palatino-bolditalic' => undef,
#    'palatino-italic' => undef,
#    'palatino-roman' => undef,
#    'symbol' => undef,
#    'zapfchancery-mediumitalic' => undef,
#    'zapfdingbats' => undef,
#   );

sub enscript {
    my($top, %args) = @_;

    my $t = $top->Toplevel;

    my $media = $args{-media} || 'A4';
    die "Unknown media $media" if !exists $media{$media};
    my %media_desc = %{$media{$media}};

    my $width  = $args{-width}  || $media_desc{Width};
    my $height = $args{-height} || $media_desc{Height};

    my $c = $t->Canvas(-width => $width, -height => $height);
    $t->withdraw;

    my($llx, $lly, $urx, $ury) = @{$args{-bbox}} if exists $args{-bbox};

    $llx = $args{-llx} || $media_desc{LLX};
    $lly = $args{-lly} || $media_desc{LLY};
    $urx = $args{-urx} || $media_desc{URX};
    $ury = $args{-ury} || $media_desc{URY};

    my $uly = $height - $ury; # XXX unsure
    my $lry = $height - $lly;
    
    my $y = $uly;
    my $font = new Tk::Font
      ($t, postscript_to_x11_font($args{-font} || 'Courier-12'));

    my $output = $args{-output} || "/tmp/enscript.%d.ps";

    my $page = 0;
    my $line;

    my $ps_output_sub = sub {
	$c->update;
	$c->postscript(-file => sprintf($output, $page),
		       -pagewidth => $width,
		       -pageheight => $height,
		       -width => $width,
		       -height => $height);
	$y = $uly;
	$page++;
	$c->delete('all');
    };

    if (exists $args{-file}) {
	open(F, $args{-file}) or die "Can't open $args{-file}: $!";
	local($/) = undef;
	$args{-text} = <F>;
	close F;
    }

    my $try_again = 0;
    foreach $line (split(/\n/, $args{-text})) {
	$line = expand($line);
	my $i = $c->createText($llx, $y,
			       -width => $urx-$llx,
			       -text => $line, -anchor => 'nw',
			       -font => $font, 
			      );
	$y = ($c->bbox($i))[3];
	if ($y > $lry && !$try_again) {
	    $c->delete($i);
	    $ps_output_sub->();
	    $try_again++;
	    redo;
	}
	$try_again = 0;
    }

    $ps_output_sub->();
    $c->destroy;

    ($output, $page-1); # gibt Output-Dateiname und Anzahl der Seiten zurck
}

sub parse_cfg {
    my $cfg_file = shift;
    if (!defined $cfg_file) {
	$cfg_file = Tk->findINC('enscript.cfg');
    }

    %media = ();
    %postscript_to_x11_font = ();

    open(CFG, $cfg_file)
      or die "Can't open config file <$cfg_file>: $!";
    while(<CFG>) {
	s/\s*\#.*//;
	next if /^\s*$/;
	if (/^\s*Media:\s*(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)/) {
	    $media{$1} = {Width  => $2,
			  Height => $3,
			  LLX    => $4,
			  LLY    => $5,
			  URX    => $6,
			  URY    => $7};
	} elsif (/^\s*FontMap:\s*(\S+)\s+(.*)/) {
	    $postscript_to_x11_font{$1} = $2;
	} else {
	    die "Can't parse $_";
	}
    }
    close CFG;
}

sub postscript_to_x11_font {
    my($psfont) = @_;
    my $x11font;
    if ($psfont !~ /^(.*)-(\d+)$/) {
	die "Can't parse postscript font $psfont";
    }
    my($font, $size) = (lc($1), $2);
    if (!defined $postscript_to_x11_font{$font}) {
	die "No X11 font for $font defined";
    }
    $x11font = sprintf($postscript_to_x11_font{$font}, $size*10);
    $x11font;
}

1;

=head1 NAME

Tk::Enscript - a text-to-postscript converter using Tk::Canvas

=head1 SYNOPSIS

    use Tk::Enscript;

    enscript($top,
             -text   => $text,
             -media  => 'A4',
             -output => "/tmp/bla.%d.ps",
            );

=head1 DESCRIPTION

=head1 AUTHOR

Slaven Rezic <eserte@cs.tu-berlin.de>

=head1 COPYRIGHT

Copyright (c) 1998 Slaven Rezic. All rights reserved.
This module is free software; you can redistribute it and/or modify
it under the same terms as Perl itself.

=head1 SEE ALSO

enscript(1), a2ps(1), Tk::Canvas(3)

=cut

